from __future__ import annotations

from pathlib import Path

import geopandas as gpd
import networkx as nx
import numpy as np
import pandas as pd
import pytest
import shapely

from swmmanywhere import metric_utilities as mu
from swmmanywhere.graph_utilities import load_graph
from swmmanywhere.parameters import MetricEvaluation


def assert_close(a: float, b: float, rtol: float = 1e-3) -> None:
    """Assert that two floats are close."""
    assert np.isclose(a, b, rtol=rtol).all()


@pytest.fixture
def subs():
    """Get a GeoDataFrame of subcatchments."""
    subs = [
        shapely.Polygon(
            [
                (700262, 5709928),
                (700262, 5709883),
                (700351, 5709883),
                (700351, 5709906),
                (700306, 5709906),
                (700306, 5709928),
                (700262, 5709928),
            ]
        ),
        shapely.Polygon(
            [
                (700306, 5709928),
                (700284, 5709928),
                (700284, 5709950),
                (700374, 5709950),
                (700374, 5709906),
                (700351, 5709906),
                (700306, 5709906),
                (700306, 5709928),
            ]
        ),
        shapely.Polygon(
            [
                (700351, 5709883),
                (700351, 5709906),
                (700374, 5709906),
                (700374, 5709883),
                (700396, 5709883),
                (700396, 5709816),
                (700329, 5709816),
                (700329, 5709838),
                (700329, 5709883),
                (700351, 5709883),
            ]
        ),
    ]

    subs = gpd.GeoDataFrame(
        data={"id": [107733, 1696030874, 6277683849]}, geometry=subs, crs="EPSG:32630"
    )
    return subs


@pytest.fixture
def results():
    """Get a DataFrame of results."""
    results = pd.DataFrame(
        [
            {
                "id": 4253560,
                "variable": "flow",
                "value": 10,
                "date": pd.to_datetime("2021-01-01 00:00:00"),
            },
            {
                "id": 4253560,
                "variable": "flow",
                "value": 5,
                "date": pd.to_datetime("2021-01-01 00:00:05"),
            },
            {
                "id": 1696030874,
                "variable": "flooding",
                "value": 4.5,
                "date": pd.to_datetime("2021-01-01 00:00:00"),
            },
            {
                "id": 770549936,
                "variable": "flooding",
                "value": 5,
                "date": pd.to_datetime("2021-01-01 00:00:00"),
            },
            {
                "id": 107736,
                "variable": "flooding",
                "value": 10,
                "date": pd.to_datetime("2021-01-01 00:00:00"),
            },
            {
                "id": 107733,
                "variable": "flooding",
                "value": 1,
                "date": pd.to_datetime("2021-01-01 00:00:00"),
            },
            {
                "id": 107737,
                "variable": "flooding",
                "value": 2,
                "date": pd.to_datetime("2021-01-01 00:00:00"),
            },
            {
                "id": 1696030874,
                "variable": "flooding",
                "value": 0,
                "date": pd.to_datetime("2021-01-01 00:00:05"),
            },
            {
                "id": 770549936,
                "variable": "flooding",
                "value": 5,
                "date": pd.to_datetime("2021-01-01 00:00:05"),
            },
            {
                "id": 107736,
                "variable": "flooding",
                "value": 15,
                "date": pd.to_datetime("2021-01-01 00:00:05"),
            },
            {
                "id": 107733,
                "variable": "flooding",
                "value": 2,
                "date": pd.to_datetime("2021-01-01 00:00:05"),
            },
            {
                "id": 107737,
                "variable": "flooding",
                "value": 2,
                "date": pd.to_datetime("2021-01-01 00:00:05"),
            },
        ]
    )
    return results


def test_bias_flood_depth():
    """Test the bias_flood_depth metric."""
    # Create synthetic and real data
    synthetic_results = pd.DataFrame(
        {
            "id": ["obj1", "obj1", "obj2", "obj2"],
            "value": [10, 20, 5, 2],
            "variable": "flooding",
            "date": pd.to_datetime(
                [
                    "2021-01-01 00:00:00",
                    "2021-01-01 00:05:00",
                    "2021-01-01 00:00:00",
                    "2021-01-01 00:05:00",
                ]
            ),
        }
    )
    real_results = pd.DataFrame(
        {
            "id": ["obj1", "obj1", "obj2", "obj2"],
            "value": [15, 25, 10, 20],
            "variable": "flooding",
            "date": pd.to_datetime(
                [
                    "2021-01-01 00:00:00",
                    "2021-01-01 00:05:00",
                    "2021-01-01 00:00:00",
                    "2021-01-01 00:05:00",
                ]
            ),
        }
    )
    synthetic_subs = pd.DataFrame(
        {
            "impervious_area": [100, 200],
        }
    )
    real_subs = pd.DataFrame(
        {
            "impervious_area": [150, 250],
        }
    )

    # Run the metric
    val = mu.metrics.bias_flood_depth(
        synthetic_results=synthetic_results,
        real_results=real_results,
        synthetic_subs=synthetic_subs,
        real_subs=real_subs,
    )
    assert_close(val, -0.2952)


def test_kstest_betweenness():
    """Test the kstest_betweenness metric."""
    G = load_graph(Path(__file__).parent / "test_data" / "graph_topo_derived.json")
    val = mu.metrics.kstest_betweenness(synthetic_G=G, real_G=G)
    assert val == 0.0

    G_ = G.copy()
    G_.remove_node(list(G.nodes)[0])
    val = mu.metrics.kstest_betweenness(synthetic_G=G_, real_G=G)
    assert_close(val, 0.2862)


def test_kstest_edge_betweenness():
    """Test the kstest_betweenness metric."""
    G = load_graph(Path(__file__).parent / "test_data" / "graph_topo_derived.json")
    val = mu.metrics.kstest_edge_betweenness(synthetic_G=G, real_G=G)
    assert val == 0.0

    G_ = G.copy()
    G_.remove_node(list(G.nodes)[0])
    val = mu.metrics.kstest_edge_betweenness(synthetic_G=G_, real_G=G)
    assert_close(val, 0.38995)


def test_best_outfall_match(subs):
    """Test the best_outfall_match and ks_betweenness."""
    G = load_graph(Path(__file__).parent / "test_data" / "graph_topo_derived.json")

    sg, outfall = mu.best_outfall_match(synthetic_G=G, real_subs=subs)
    outfalls = nx.get_node_attributes(sg, "outfall")
    assert len(set(outfalls.values())) == 1
    assert outfall == 12354833


def test_nse():
    """Test the nse metric."""
    val = mu.nse(y=np.array([1, 2, 3, 4, 5]), yhat=np.array([1, 2, 3, 4, 5]))
    assert val == 1.0

    val = mu.nse(y=np.array([1, 2, 3, 4, 5]), yhat=np.array([3, 3, 3, 3, 3]))
    assert val == 0.0


def test_kge():
    """Test the kge metric."""
    val = mu.kge(y=np.array([1, 2, 3, 4, 5]), yhat=np.array([1, 2, 3, 4, 5]))
    assert_close(val, 1.0)

    val = mu.kge(y=np.array([1, 2, 3, 4, 5]), yhat=np.array([3, 3, 3, 3, 3]))
    assert_close(val, (1 - 2**0.5))


def test_inf():
    """Test metrics handling of invalid coefficients."""
    val = mu.kge(y=np.array([3, 3, 3, 3, 3]), yhat=np.array([1, 2, 3, 4, 5]))
    assert val == np.inf

    val = mu.nse(y=np.array([3, 3, 3, 3, 3]), yhat=np.array([1, 2, 3, 4, 5]))
    assert val == np.inf

    val = mu.relerror(y=np.array([-3, -3, 0, 3, 3]), yhat=np.array([1, 2, 3, 4, 5]))
    assert val == np.inf


def test_relerror_different_length():
    """Test the relerror metric with different length arrays."""
    val = mu.relerror(y=np.array([1, 2, 3, 4, 5, 6]), yhat=np.array([1]))
    assert_close(val, (1 - 3.5) / 3.5)


def test_outfall_nse_flow(subs):
    """Test the outfall_nse_flow metric."""
    # Load data
    G = load_graph(Path(__file__).parent / "test_data" / "graph_topo_derived.json")

    # Mock results
    results = pd.DataFrame(
        [
            {
                "id": 4253560,
                "variable": "flow",
                "value": 10,
                "date": pd.to_datetime("2021-01-01").date(),
            },
            {
                "id": "",
                "variable": "flow",
                "value": 5,
                "date": pd.to_datetime("2021-01-01").date(),
            },
            {
                "id": 4253560,
                "variable": "flow",
                "value": 5,
                "date": pd.to_datetime("2021-01-01 00:00:05"),
            },
            {
                "id": "",
                "variable": "flow",
                "value": 2,
                "date": pd.to_datetime("2021-01-01 00:00:05"),
            },
        ]
    )

    # Calculate NSE (perfect results)
    val = mu.metrics.outfall_nse_flow(
        synthetic_G=G,
        synthetic_subs=None,
        synthetic_results=results,
        real_G=G,
        real_results=results,
        real_subs=subs,
        metric_evaluation=None,
    )
    assert val == 1.0

    # Calculate NSE (mean results)
    results_ = results.copy()
    results_.loc[[0, 2], "value"] = 7.5
    val = mu.metrics.outfall_nse_flow(
        synthetic_G=G,
        synthetic_subs=None,
        synthetic_results=results_,
        real_G=G,
        real_results=results,
        real_subs=subs,
        metric_evaluation=None,
    )
    assert val == 0.0

    # Change the graph
    G_ = G.copy()
    new_outfall = list(G_.in_edges(12354833))[0][0]
    nx.set_node_attributes(G_, new_outfall, "outfall")
    G_.remove_node(12354833)
    results_.loc[results_.id == 4253560, "id"] = 725226531

    # Calculate NSE (mean results)
    val = mu.metrics.outfall_nse_flow(
        synthetic_G=G_,
        synthetic_subs=None,
        synthetic_results=results_,
        real_G=G,
        real_results=results,
        real_subs=subs,
        metric_evaluation=None,
    )
    assert val == 0.0

    # Test time interpolation
    results_.loc[2, "date"] = pd.to_datetime("2021-01-01 00:00:10")
    results_.loc[[0, 2], "value"] = [0, 30]
    val = mu.metrics.outfall_nse_flow(
        synthetic_G=G_,
        synthetic_subs=None,
        synthetic_results=results_,
        real_G=G,
        real_results=results,
        real_subs=subs,
        metric_evaluation=None,
    )
    assert val == -15.0


def test_outfall_nse_flooding(subs):
    """Test the outfall_nse_flow metric."""
    # Load data
    G = load_graph(Path(__file__).parent / "test_data" / "graph_topo_derived.json")

    # Mock results
    results = pd.DataFrame(
        [
            {
                "id": 4253560,
                "variable": "flow",
                "value": 10,
                "date": pd.to_datetime("2021-01-01 00:00:00"),
            },
            {
                "id": 4253560,
                "variable": "flow",
                "value": 5,
                "date": pd.to_datetime("2021-01-01 00:00:05"),
            },
            {
                "id": 25472468,
                "variable": "flooding",
                "value": 4.5,
                "date": pd.to_datetime("2021-01-01 00:00:00"),
            },
            {
                "id": 770549936,
                "variable": "flooding",
                "value": 5,
                "date": pd.to_datetime("2021-01-01 00:00:00"),
            },
            {
                "id": 109753,
                "variable": "flooding",
                "value": 10,
                "date": pd.to_datetime("2021-01-01 00:00:00"),
            },
            {
                "id": 25472468,
                "variable": "flooding",
                "value": 0,
                "date": pd.to_datetime("2021-01-01 00:00:05"),
            },
            {
                "id": 770549936,
                "variable": "flooding",
                "value": 5,
                "date": pd.to_datetime("2021-01-01 00:00:05"),
            },
            {
                "id": 109753,
                "variable": "flooding",
                "value": 15,
                "date": pd.to_datetime("2021-01-01 00:00:05"),
            },
        ]
    )

    # Calculate NSE (perfect results)
    val = mu.metrics.outfall_nse_flooding(
        synthetic_G=G,
        synthetic_subs=None,
        synthetic_results=results,
        real_G=G,
        real_results=results,
        real_subs=subs,
        metric_evaluation=None,
    )
    assert val == 1.0

    # Calculate NSE (mean results)
    results_ = results.copy()
    results_.loc[results_.id.isin([770549936, 25472468]), "value"] = [14.5 / 4] * 4
    val = mu.metrics.outfall_nse_flooding(
        synthetic_G=G,
        synthetic_subs=None,
        synthetic_results=results_,
        real_G=G,
        real_results=results,
        real_subs=subs,
        metric_evaluation=None,
    )
    assert val == 0.0

    # Change the graph
    G_ = G.copy()
    new_outfall = list(G_.in_edges(12354833))[0][0]
    nx.set_node_attributes(
        G_,
        {x: new_outfall for x, d in G_.nodes(data=True) if d["outfall"] == 12354833},
        "outfall",
    )
    G_.remove_node(12354833)

    # Calculate NSE (mean results)
    val = mu.metrics.outfall_nse_flooding(
        synthetic_G=G_,
        synthetic_subs=None,
        synthetic_results=results_,
        real_G=G,
        real_results=results,
        real_subs=subs,
        metric_evaluation=None,
    )
    assert val == 0.0

    # Test time interpolation
    results_.loc[results_.date == pd.to_datetime("2021-01-01 00:00:05"), "date"] = (
        pd.to_datetime("2021-01-01 00:00:10")
    )

    val = mu.metrics.outfall_nse_flooding(
        synthetic_G=G_,
        synthetic_results=results_,
        synthetic_subs=None,
        real_G=G,
        real_results=results,
        real_subs=subs,
        metric_evaluation=None,
    )
    assert val == 0.0


def test_design_params(subs):
    """Test the design param related metrics."""
    G = load_graph(Path(__file__).parent / "test_data" / "graph_topo_derived.json")
    nx.set_edge_attributes(G, 0.15, "diameter")

    # Mock results (only needed for dominant outfall)
    results = pd.DataFrame(
        [
            {
                "id": 4253560,
                "variable": "flow",
                "value": 10,
                "date": pd.to_datetime("2021-01-01 00:00:00"),
            },
            {
                "id": 4253560,
                "variable": "flow",
                "value": 5,
                "date": pd.to_datetime("2021-01-01 00:00:05"),
            },
        ]
    )

    # Target results
    design_results = {
        "outfall_kstest_diameters": 0.0625,
        "outfall_relerror_diameter": 0.0625,
        "outfall_relerror_length": -0.15088965,
        "outfall_relerror_nmanholes": -0.05,
        "outfall_relerror_npipes": -0.15789473,
    }

    # Iterate for G = G, i.e., perfect results
    metrics = mu.iterate_metrics(
        synthetic_G=G,
        synthetic_subs=None,
        synthetic_results=None,
        real_G=G,
        real_subs=subs,
        real_results=results,
        metric_list=design_results.keys(),
        metric_evaluation=MetricEvaluation(),
    )
    for metric, val in metrics.items():
        assert metric in design_results
        assert np.isclose(val, 0)

    # edit the graph for target results
    G_ = G.copy()
    G_.remove_node(list(G.nodes)[0])
    G_.edges[list(G_.edges)[0]]["diameter"] = 0.3

    metrics = mu.iterate_metrics(
        synthetic_G=G_,
        synthetic_subs=None,
        synthetic_results=None,
        real_G=G,
        real_subs=subs,
        real_results=results,
        metric_list=design_results.keys(),
        metric_evaluation=MetricEvaluation(),
    )

    for metric, val in metrics.items():
        assert metric in design_results
        assert np.isclose(val, design_results[metric]), metric


def test_netcomp_iterate():
    """Test the netcomp metrics and iterate_metrics."""
    netcomp_results = {
        "nc_deltacon0": 0.00129408,
        "nc_laplacian_dist": 155.428234,
        "nc_laplacian_norm_dist": 0.752901,
        "nc_adjacency_dist": 226.22945,
        "nc_resistance_distance": 8.098548,
        "nc_vertex_edge_distance": 0.132075,
    }
    G = load_graph(Path(__file__).parent / "test_data" / "graph_topo_derived.json")
    metrics = mu.iterate_metrics(
        synthetic_G=G,
        synthetic_subs=None,
        synthetic_results=None,
        real_G=G,
        real_subs=None,
        real_results=None,
        metric_list=netcomp_results.keys(),
        metric_evaluation=MetricEvaluation(),
    )
    for metric, val in metrics.items():
        assert metric in netcomp_results
        assert np.isclose(val, 0)

    G_ = load_graph(Path(__file__).parent / "test_data" / "street_graph.json")
    metrics = mu.iterate_metrics(
        synthetic_G=G_,
        synthetic_subs=None,
        synthetic_results=None,
        real_G=G,
        real_subs=None,
        real_results=None,
        metric_list=netcomp_results.keys(),
        metric_evaluation=MetricEvaluation(),
    )
    for metric, val in metrics.items():
        assert metric in netcomp_results
        assert np.isclose(val, netcomp_results[metric])


def test_subcatchment_nse_flooding(subs, results):
    """Test the outfall_nse_flow metric."""
    # Load data
    G = load_graph(Path(__file__).parent / "test_data" / "graph_topo_derived.json")

    # Calculate NSE (perfect results)
    val = mu.metrics.subcatchment_nse_flooding(
        synthetic_G=G,
        synthetic_subs=None,
        real_G=G,
        synthetic_results=results,
        real_results=results,
        real_subs=subs,
        metric_evaluation=None,
    )
    assert val == 1.0

    # Calculate NSE (remapped node)

    G_ = G.copy()
    # Create a mapping from the old name to the new name
    mapping = {1696030874: "new_name", 107737: "new_name2"}

    # Rename the node
    G_ = nx.relabel_nodes(G_, mapping)

    results_ = results.copy()
    results_.id = results_.id.replace(mapping)

    val = mu.metrics.subcatchment_nse_flooding(
        synthetic_G=G_,
        synthetic_subs=None,
        synthetic_results=results_,
        real_G=G,
        real_results=results,
        real_subs=subs,
        metric_evaluation=None,
    )
    assert val == 1.0

    # Test gridded
    val = mu.metrics.grid_nse_flooding(
        synthetic_G=G_,
        synthetic_subs=None,
        synthetic_results=results_,
        real_G=G,
        real_results=results,
        real_subs=subs,
        metric_evaluation=MetricEvaluation(),
    )
    assert val == 1.0


def test_create_grid():
    """Test the create_grid function."""
    grid = mu.create_grid((0, 0, 1, 1), 1 / 3 - 0.001)
    assert grid.shape[0] == 16
    assert set(grid.columns) == {"sub_id", "geometry"}


def test_restirctions():
    """Test the restriction register by generating an invalid metric."""
    # Invalid because length can't be calculated at grid scale
    with pytest.raises(ValueError):
        mu.metric_factory("grid_relerror_length")

    # Invalid because nmanholes can't be evaluated with nse
    with pytest.raises(ValueError):
        mu.metric_factory("outfall_nse_nmanholes")


def test_nodes_to_subs(subs):
    """Test the nodes_to_subs function."""
    G = load_graph(Path(__file__).parent / "test_data" / "graph_topo_derived.json")
    nodes = mu.nodes_to_subs(G, subs)
    nodes = set([(r.id, r.sub_id) for _, r in nodes.iterrows()])
    nodes_ = {
        (107733, 107733),
        (107736, 1696030874),
        (107737, 107733),
        (32925453, 6277683849),
        (770549936, 1696030874),
        (6277683849, 6277683849),
    }
    assert nodes == nodes_


def test_align_by_shape_and_median_coef(subs, results):
    """Test the align_by_shape and median_coef_by_group function."""
    G = load_graph(Path(__file__).parent / "test_data" / "graph_topo_derived.json")
    # Align the grid
    aligned = mu.align_by_shape("flooding", results, results, subs, G, G)
    assert aligned.shape[0] == 4
    assert (aligned.value_real == aligned.value_syn).all()

    # Test interp
    results_ = results.copy()
    results_ = pd.concat(
        [
            results_,
            results_.replace(
                {
                    "date": {
                        pd.to_datetime("2021-01-01 00:00:00"): pd.to_datetime(
                            "2021-01-01 00:00:03"
                        ),
                        pd.to_datetime("2021-01-01 00:00:05"): pd.to_datetime(
                            "2021-01-01 00:00:02"
                        ),
                    }
                }
            ),
        ]
    )

    aligned = mu.align_by_shape("flooding", results, results_, subs, G, G)
    assert aligned.shape[0] == 8
    assert (
        aligned.loc[aligned.date.isin(results.date)].value_real
        == aligned.loc[aligned.date.isin(results.date)].value_syn
    ).all()
    assert_close(
        aligned.loc[[2, 3, 4, 5], "value_syn"], [3.3333, 16.6666, 3.6666, 18.3333]
    )

    # Test median coef
    meds = mu.median_coef_by_group(aligned, "sub_id", mu.nse)
    assert_close(meds, 0.1111)


def test_apply_warmup():
    """Test the apply_warmup function."""
    # Create a DataFrame with a date range
    df = pd.concat(
        [
            pd.DataFrame(
                {
                    "date": pd.date_range(
                        start="2021-01-01", periods=100, freq=f"{ds}s"
                    ),
                    "value": range(100),
                    "variable": [variable] * 100,
                }
            )
            for ds, variable in zip([2, 3], ["flow", "flooding"])
        ],
        axis=0,
    )

    # Apply warmup
    df_warmup = mu.apply_warmup(df, 0.5)

    # Check the result
    assert df_warmup.shape[0] == 75
    assert df_warmup.date.min() == pd.to_datetime("2021-01-01 00:02:30")



# %% [markdown]
# # Custom data demo
# Note - this script can also be opened in interactive Python if you wanted to
# play around. On the GitHub it is in [docs/notebooks](./notebooks/custom_data_demo.py).
# To run this on your local machine, you will need to install the optional dependencies
# for `doc`:
#
# `pip install swmmanywhere[doc]`
#
# %% [markdown]
# ## Introduction
# This script demonstrates how to use `swmmanywhere` when you have custom data.
#
# Since this is a notebook, we will define [`config`](https://imperialcollegelondon.github.io/SWMManywhere/config_guide/)
# as a dictionary rather than a `yaml` file, but the same principles apply.
#
# ## Initial setup
#
# We will use the same example as the [extended demo](./notebooks/extended_demo.py), but
# with a custom elevation dataset. Let's start by rerunning it.
# %%
# Imports
from __future__ import annotations

import tempfile
from pathlib import Path

import folium
import geopandas as gpd

from swmmanywhere.logging import set_verbose
from swmmanywhere.swmmanywhere import swmmanywhere
from swmmanywhere.utilities import plot_map

# Create temporary directory
temp_dir = tempfile.TemporaryDirectory(dir=".")
base_dir = Path(temp_dir.name)

# Define minimum viable config
bbox = (1.52740, 42.50524, 1.54273, 42.51259)
config = {
    "base_dir": base_dir,
    "project": "my_first_swmm",
    "bbox": bbox,
    "run_settings": {"duration": 3600},
    "parameter_overrides": {
        "topology_derivation": {
            "allowable_networks": ["drive"],
            "omit_edges": ["bridge"],
        },
        "outfall_derivation": {
            "outfall_length": 5,
            "river_buffer_distance": 30,
        },
    },
}
set_verbose(True)  # Set verbosity

# Run SWMManywhere
outputs = swmmanywhere(config)
model_dir = outputs[0].parent

# %% [markdown]
# ## Plotting output
#
# Now we can plot the output. To highlight the differences in the supplied data that we
# are about to demonstrate, we also plot the subbasins.
# %%
m = plot_map(model_dir)
subbasins = gpd.read_file(model_dir / "subbasins.geoparquet")
folium.GeoJson(subbasins, fill_opacity=0, color="blue", weight=2).add_to(m)
m


# %% [markdown]
# ## Supply custom elevation data
#
# To keep things simple, we will just download some elevation data for the same area
# and perturb it, though in practice you are likely to use some higher resolution
# or more accurate data.
#
# You don't need to worry about your files lining up perfectly (though of course if they
# do not overlap at all then you will run into problems).
# %%

# Import NASADEM downloader and reprojection tools
from swmmanywhere.geospatial_utilities import (  # noqa: E402
    get_utm_epsg,
    reproject_raster,
)
from swmmanywhere.prepare_data import download_elevation  # noqa: E402

# Download and reproject the correct elevation to UTM
download_elevation(base_dir / "elevation.tif", bbox)
reproject_raster(
    get_utm_epsg(bbox[0], bbox[1]),
    base_dir / "elevation.tif",
    base_dir / "elevation_utm.tif",
)

# Flip it
import numpy as np  # noqa: E402
import rasterio  # noqa: E402

with rasterio.open(base_dir / "elevation_utm.tif") as src:
    data = np.fliplr(src.read(1))
    with rasterio.open(base_dir / "fake_elevation.tif", "w", **src.profile) as dst:
        dst.write(data, 1)

# %% [markdown]
# ## Update config and run again
#
# Now we update the `elevation` entry in the `address_overrides` part of the
# `config` to point to the new elevation data, then rerun `swmmanywhere`.
# %%
# Update config
config["address_overrides"] = {
    "elevation": str(base_dir / "fake_elevation.tif"),
}

# Run again
outputs = swmmanywhere(config)
model_dir = outputs[0].parent

# %% [markdown]
# ## Plotting output
#
# This time we will include both the original (blue) and the new (red) subbasins to
# highlight the impact of flipping the elevation data.
# %%
m = plot_map(model_dir)
subbasins_new = gpd.read_file(model_dir / "subbasins.geoparquet")
folium.GeoJson(subbasins_new, fill_opacity=0, color="red", weight=2).add_to(m)
folium.GeoJson(subbasins, fill_opacity=0, color="blue", weight=2).add_to(m)
m

# %% [markdown]
#
# ## In Summary
#
# This tutorial has demonstrated how we can supply our own data in place of the existing
# files that are downloaded by SWMManywhere. This method works for any
# [filepath](https://imperialcollegelondon.github.io/SWMManywhere/reference-filepaths/),
# which are the files used by the current workflow.
#
# You can also use `address_overrides` to you add your own custom data that
# might be used by a [custom graphfcn](https://imperialcollegelondon.github.io/SWMManywhere/graphfcns_guide/)
# , for example. If you are adding new data to the SWMManywhere workflow in this way,
# you will have to access it from the `FilePaths` object with the `get_path` method.
#



"""Module for graphfcns that change subcatchments."""

from __future__ import annotations

import sys
from collections import defaultdict
from typing import Any, Dict, List

import networkx as nx
import numpy as np

from swmmanywhere import geospatial_utilities as go
from swmmanywhere import parameters, shortest_path_utils
from swmmanywhere.filepaths import FilePaths
from swmmanywhere.graph_utilities import (
    BaseGraphFunction,
    filter_streets,
    register_graphfcn,
)
from swmmanywhere.logging import logger


@register_graphfcn
class set_elevation(
    BaseGraphFunction,
    required_node_attributes=["x", "y"],
    adds_node_attributes=["surface_elevation"],
):
    """set_elevation class."""

    def __call__(self, G: nx.Graph, addresses: FilePaths, **kwargs) -> nx.Graph:
        """Set the elevation for each node.

        This function sets the elevation for each node. The elevation is
        calculated from the elevation data.

        Args:
            G (nx.Graph): A graph
            addresses (FilePaths): An FilePaths parameter object
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.Graph): A graph
        """
        G = G.copy()
        x = [d["x"] for x, d in G.nodes(data=True)]
        y = [d["y"] for x, d in G.nodes(data=True)]
        elevations = go.interpolate_points_on_raster(
            x, y, addresses.bbox_paths.elevation
        )
        elevations_dict = {id_: elev for id_, elev in zip(G.nodes, elevations)}
        nx.set_node_attributes(G, elevations_dict, "surface_elevation")
        return G


@register_graphfcn
class set_surface_slope(
    BaseGraphFunction,
    required_node_attributes=["surface_elevation"],
    adds_edge_attributes=["surface_slope"],
):
    """set_surface_slope class."""

    def __call__(self, G: nx.Graph, **kwargs) -> nx.Graph:
        """Set the surface slope for each edge.

        This function sets the surface slope for each edge. The surface slope is
        calculated from the elevation data.

        Args:
            G (nx.Graph): A graph
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.Graph): A graph
        """
        G = G.copy()
        # Compute the slope for each edge
        slope_dict = {
            (u, v, k): (
                G.nodes[u]["surface_elevation"] - G.nodes[v]["surface_elevation"]
            )
            / d["length"]
            for u, v, k, d in G.edges(data=True, keys=True)
        }

        # Set the 'surface_slope' attribute for all edges
        nx.set_edge_attributes(G, slope_dict, "surface_slope")
        return G


@register_graphfcn
class set_chahinian_slope(
    BaseGraphFunction,
    required_edge_attributes=["surface_slope"],
    adds_edge_attributes=["chahinian_slope"],
):
    """set_chahinian_slope class."""

    def __call__(self, G: nx.Graph, **kwargs) -> nx.Graph:
        """set_chahinian_slope class.

        This function sets the Chahinian slope for each edge. The Chahinian slope is
        calculated from the surface slope and weighted according to the slope
        (based on: https://doi.org/10.1016/j.compenvurbsys.2019.101370)

        Args:
            G (nx.Graph): A graph
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.Graph): A graph
        """
        G = G.copy()

        # Values where the weight of the slope can be matched to the values
        # in weights - e.g., a slope of 0.3% has 0 weight (preferred), while
        # a slope of <=-1% has a weight of 1 (not preferred)
        slope_points = [-1, 0.3, 0.7, 10]
        weights = [1, 0, 0, 1]

        # Calculate weights
        slope = nx.get_edge_attributes(G, "surface_slope")
        weights = np.interp(
            np.asarray(list(slope.values())) * 100,
            slope_points,
            weights,
            left=1,
            right=1,
        )
        nx.set_edge_attributes(G, dict(zip(slope, weights)), "chahinian_slope")

        return G


@register_graphfcn
class set_chahinian_angle(
    BaseGraphFunction,
    required_node_attributes=["x", "y"],
    adds_edge_attributes=["chahinian_angle"],
):
    """set_chahinian_angle class."""

    def __call__(self, G: nx.Graph, **kwargs) -> nx.Graph:
        """Set the Chahinian angle for each edge.

        This function sets the Chahinian angle for each edge. The Chahinian angle is
        calculated from the geometry of the edge and weighted according to the
        angle (based on: https://doi.org/10.1016/j.compenvurbsys.2019.101370)

        Args:
            G (nx.Graph): A graph
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.Graph): A graph
        """
        # TODO - in a double directed graph, not sure how meaningful this is
        # TODO could probably refactor
        G = G.copy()
        for u, v, d in G.edges(data=True):
            min_weight = float("inf")
            for node in G.successors(v):
                if node == u:
                    continue

                p1 = (G.nodes[u]["x"], G.nodes[u]["y"])
                p2 = (G.nodes[v]["x"], G.nodes[v]["y"])
                p3 = (G.nodes[node]["x"], G.nodes[node]["y"])
                angle = go.calculate_angle(p1, p2, p3)
                chahinian_weight = np.interp(
                    angle,
                    [0, 90, 135, 180, 225, 270, 360],
                    [1, 0.2, 0.7, 0, 0.7, 0.2, 1],
                )
                min_weight = min(chahinian_weight, min_weight)
            if min_weight == float("inf"):
                min_weight = 0
            d["chahinian_angle"] = min_weight
        return G


@register_graphfcn
class calculate_weights(
    BaseGraphFunction,
    required_edge_attributes=parameters.TopologyDerivation().weights,
    adds_edge_attributes=["weight"],
):
    """calculate_weights class."""

    # TODO.. I guess if someone defines their own weights, this will need
    # to change, will want an automatic way to do that...

    def __call__(
        self, G: nx.Graph, topology_derivation: parameters.TopologyDerivation, **kwargs
    ) -> nx.Graph:
        """Calculate the weights for each edge.

        This function calculates the weights for each edge. The weights are
        calculated from the edge attributes.

        Args:
            G (nx.Graph): A graph
            topology_derivation (parameters.TopologyDerivation): A TopologyDerivation
                parameter object
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.Graph): A graph
        """
        # Calculate bounds to normalise between
        bounds: Dict[Any, List[float]] = defaultdict(lambda: [np.inf, -np.inf])

        for w in topology_derivation.weights:
            bounds[w][0] = min(nx.get_edge_attributes(G, w).values())  # lower bound
            bounds[w][1] = max(nx.get_edge_attributes(G, w).values())  # upper bound

        # Avoid division by zero
        bounds = {w: [b[0], b[1]] for w, b in bounds.items() if b[0] != b[1]}

        G = G.copy()
        eps = np.finfo(float).eps
        for u, v, d in G.edges(data=True):
            total_weight = 0
            for attr, bds in bounds.items():
                # Normalise
                weight = max((d[attr] - bds[0]) / (bds[1] - bds[0]), eps)
                # Exponent
                weight = weight ** getattr(topology_derivation, f"{attr}_exponent")
                # Scaling
                weight = weight * getattr(topology_derivation, f"{attr}_scaling")
                # Sum
                total_weight += weight
            # Set
            d["weight"] = total_weight
        return G


@register_graphfcn
class derive_topology(
    BaseGraphFunction,
    required_edge_attributes=[
        "edge_type",  # 'rivers' and 'streets'
        "weight",
    ],
    adds_node_attributes=["outfall", "shortest_path"],
):
    """derive_topology class."""

    def __call__(
        self, G: nx.Graph, outfall_derivation: parameters.OutfallDerivation, **kwargs
    ) -> nx.Graph:
        """Derive the topology of a graph.

        Derives the network topology based on the weighted graph of potential
        pipe carrying edges in the graph.

        Two methods are available:
        - `separate`: The original and that assumes outfalls have already been
        narrowed down from the original plausible set. Runs a djiikstra-based
        algorithm to identify the shortest path from each node to its nearest
        outfall (weighted by the 'weight' edge value). The
        returned graph is one that only contains the edges that feature  on the
        shortest paths.
        - `withtopo`: The alternative method that assumes no narrowing of plausible
        outfalls has been performed. This method runs a Tarjan's algorithm to
        identify the spanning forest starting from a `waste` node that all
        plausible outfalls are connected to (whether via a river or directly).

        In both methods, street nodes that have no plausible route to any outfall
        are removed.

        Args:
            G (nx.Graph): A graph
            outfall_derivation (parameters.OutfallDerivation): An OutfallDerivation
                parameter object
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.Graph): A graph
        """
        G = G.copy()

        visited: set = set()

        # Increase recursion limit to allow to iterate over the entire graph
        # Seems to be the quickest way to identify which nodes have a path to
        # the outfall
        original_limit = sys.getrecursionlimit()
        sys.setrecursionlimit(max(original_limit, len(G.nodes)))

        # Identify outfalls
        if outfall_derivation.method == "withtopo":
            visited = set(nx.ancestors(G, "waste")) | {"waste"}

            # Remove nodes not reachable from waste
            G.remove_nodes_from(set(G.nodes) - visited)

            # Run shorted path
            G = shortest_path_utils.tarjans_pq(G, "waste")

            G = filter_streets(G)
        else:
            outfalls = [
                u for u, v, d in G.edges(data=True) if d["edge_type"] == "outfall"
            ]
            visited = set(outfalls)
            for outfall in outfalls:
                visited = visited | set(nx.ancestors(G, outfall))

            G.remove_nodes_from(set(G.nodes) - visited)
            G = filter_streets(G)

            # Check for negative cycles
            if nx.negative_edge_cycle(G, weight="weight"):
                logger.warning("Graph contains negative cycle")

            G = shortest_path_utils.dijkstra_pq(G, outfalls)

        # Reset recursion limit
        sys.setrecursionlimit(original_limit)

        # Log total weight
        total_weight = sum([d["weight"] for u, v, d in G.edges(data=True)])
        logger.info(f"Total graph weight {total_weight}.")

        return G



"""Debugging the derive_rc_alt function.

This is an alternative function for derive_rc in geospatial_utilities/derive_rc
it is used to double check that they are performing correctly. It may also be
more computationally efficient.
"""

from __future__ import annotations

import geopandas as gpd
import pandas as pd
import shapely


def derive_rc_alt(
    subcatchments: gpd.GeoDataFrame,
    building_footprints: gpd.GeoDataFrame,
    streetcover: gpd.GeoDataFrame,
) -> gpd.GeoDataFrame:
    """Derive the Runoff Coefficient (RC) of each subcatchment (alt)."""
    building_footprints = building_footprints.copy()
    streetcover = streetcover.copy()
    subcatchments = subcatchments.copy()

    # Derive impervious without overlaps
    intersecting_geom_sidx, intersecting_geom_bidx = building_footprints.sindex.query(
        streetcover.geometry, predicate="intersects"
    )
    intersecting_building_geom = building_footprints.iloc[
        intersecting_geom_bidx
    ].geometry
    intersecting_street_geom = streetcover.iloc[intersecting_geom_sidx].geometry

    if intersecting_geom_sidx.shape[0] > 0:
        unified_geom = intersecting_building_geom.unary_union.union(
            intersecting_street_geom.unary_union
        )
        unified_geom = gpd.GeoDataFrame(
            geometry=[unified_geom], crs=building_footprints.crs
        )
    else:
        unified_geom = gpd.GeoDataFrame(geometry=[], crs=building_footprints.crs)
    building_footprints = building_footprints.drop(
        building_footprints.index[intersecting_geom_bidx], axis=0
    )
    streetcover = streetcover.drop(streetcover.index[intersecting_geom_sidx], axis=0)

    new_geoms = [
        g for g in [unified_geom, building_footprints, streetcover] if g.shape[0] > 0
    ]
    # Create the "unified" impervious geometries
    impervious = gpd.GeoDataFrame(pd.concat(new_geoms), crs=building_footprints.crs)
    subcat_tree = subcatchments.sindex
    bf_pidx, sb_pidx = subcat_tree.query(impervious.geometry, predicate="intersects")
    sb_idx = subcatchments.iloc[sb_pidx].index

    # Calculate impervious area and runoff coefficient (rc)
    subcatchments["impervious_area"] = 0.0

    # Calculate all intersection-impervious geometries
    intersection_area = shapely.intersection(
        subcatchments.iloc[sb_pidx].geometry.to_numpy(),
        impervious.iloc[bf_pidx].geometry.to_numpy(),
    )

    # Indicate which catchment each intersection is part of
    intersections = pd.DataFrame(
        [
            {"sb_idx": ix, "impervious_geometry": ia}
            for ix, ia in zip(sb_idx, intersection_area)
        ]
    )

    # Aggregate by catchment
    areas = (
        intersections.groupby("sb_idx")
        .apply(shapely.ops.unary_union)
        .apply(shapely.area)
    )

    # Store as impervious area in subcatchments
    subcatchments["impervious_area"] = 0
    subcatchments.loc[areas.index, "impervious_area"] = areas
    subcatchments["rc"] = (
        subcatchments["impervious_area"] / subcatchments.geometry.area * 100
    )
    return subcatchments



"""Integration test for the swmmanywhere command-line interface."""

from __future__ import annotations

import sys
import tempfile
from pathlib import Path

import yaml

from swmmanywhere import __main__
from swmmanywhere.logging import logger


def test_swmmanywhere_cli(tmp_path):
    """Test that the CLI can successfully run with an actual configuration."""
    base_dir = Path(tmp_path)

    # Define minimum viable config
    config = {
        "base_dir": str(base_dir),
        "project": "my_first_swmm",
        "bbox": [1.52740, 42.50524, 1.54273, 42.51259],
    }

    config_path = base_dir / "config.yml"
    with config_path.open("w") as config_file:
        yaml.dump(config, config_file)

    # Mock sys.argv to simulate command-line arguments
    sys.argv = [
        "swmmanywhere",
        "--config_path",
        str(config_path),
        "--verbose",
        "True",
    ]

    expected = b"No real network provided, returning SWMM .inp file."

    with tempfile.NamedTemporaryFile(
        suffix=".log", mode="w+b", delete=False
    ) as temp_file:
        fid = Path(temp_file.name)
        logger.add(fid)

        # Run the CLI entry point
        __main__.run()

        # Capture the output
        assert expected in temp_file.read()
        logger.remove()



"""Copy the test data files to a given directory."""

from __future__ import annotations

from pathlib import Path
from shutil import copyfile

from swmmanywhere.geospatial_utilities import graph_to_geojson
from swmmanywhere.graph_utilities import load_graph


def copy_test_data(fid: Path):
    """Copy the test data files to a given directory.

    Args:
        fid (Path): Directory to copy the test data files to.
    """
    if not fid.exists():
        raise FileNotFoundError(f"Directory does not exist: {fid}")

    defs_dir = Path(__file__).parent
    files = [
        "bellinge_small.inp",
        "bellinge_small_graph.json",
        "bellinge_small_subcatchments.geojson",
        "storm.dat",
    ]
    for filename in files:
        copyfile(defs_dir / filename, fid / filename)

    G = load_graph(fid / "bellinge_small_graph.json")
    graph_to_geojson(G, fid / "nodes.geojson", fid / "edges.geojson", G.graph["crs"])



# -*- coding: utf-8 -*-
"""Created on 2024-01-26.

@author: Barney
"""

from __future__ import annotations

import os
from pathlib import Path
from tempfile import NamedTemporaryFile

from swmmanywhere.logging import logger, verbose


def test_logger():
    """Test logger."""
    os.environ["SWMMANYWHERE_VERBOSE"] = "true"
    assert logger is not None
    logger.test_logger()
    logger.debug("This is a debug message.")
    logger.info("This is an info message.")
    logger.warning("This is a warning message.")
    logger.error("This is an error message.")
    logger.critical("This is a critical message.")
    with NamedTemporaryFile(suffix=".log", mode="w+b", delete=False) as temp_file:
        fid = Path(temp_file.name)
        logger.add(fid)
        logger.test_logger()
        assert temp_file.read() != b""
        logger.remove()
    fid.unlink()
    os.environ["SWMMANYWHERE_VERBOSE"] = "false"


def test_logger_disable():
    """Test the disable function."""
    with NamedTemporaryFile(suffix=".log", mode="w+b", delete=False) as temp_file:
        fid = Path(temp_file.name)
        os.environ["SWMMANYWHERE_VERBOSE"] = "false"
        logger.add(fid)
        logger.test_logger()
        assert temp_file.read() == b""
        logger.remove()
    fid.unlink()


def test_logger_reimport():
    """Reimport logger to check that changes from disable are persistent."""
    from swmmanywhere.logging import logger

    with NamedTemporaryFile(suffix=".log", mode="w+b", delete=False) as temp_file:
        fid = Path(temp_file.name)
        logger.add(fid)
        logger.test_logger()
        assert temp_file.read() == b""
        logger.remove()
    fid.unlink()


def test_logger_again():
    """Test the logger after these changes to make sure still works."""
    os.environ["SWMMANYWHERE_VERBOSE"] = "true"
    with NamedTemporaryFile(suffix=".log", mode="w+b", delete=False) as temp_file:
        fid = Path(temp_file.name)
        logger.add(fid)
        logger.test_logger()
        assert temp_file.read() != b""
        logger.remove()
    fid.unlink()
    os.environ["SWMMANYWHERE_VERBOSE"] = "false"


def test_verbose():
    """Test the verbose function."""
    os.environ["SWMMANYWHERE_VERBOSE"] = "true"
    assert verbose()

    os.environ["SWMMANYWHERE_VERBOSE"] = "false"
    assert not verbose()

    del os.environ["SWMMANYWHERE_VERBOSE"]
    assert not verbose()



"""Module for graph functions."""

from __future__ import annotations

# Importing the graphfcns modules to register them
from . import (
    design_graphfcns,  # noqa: F401
    network_cleaning_graphfcns,  # noqa: F401
    outfall_graphfcns,  # noqa: F401
    subcatchment_graphfcns,  # noqa: F401
    topology_graphfcns,  # noqa: F401
)



"""Module for graphfcns that change the plausible pipe location network."""

from __future__ import annotations

import re
from typing import Any, Hashable, cast

import geopandas as gpd
import networkx as nx
import osmnx as ox
import pandas as pd
import shapely

from swmmanywhere import geospatial_utilities as go
from swmmanywhere import parameters
from swmmanywhere.filepaths import FilePaths
from swmmanywhere.graph_utilities import BaseGraphFunction, register_graphfcn


def get_osmid_id(data: dict) -> Hashable:
    """Get the ID of an edge."""
    id_ = data.get("osmid", data.get("id"))
    if isinstance(id_, list):
        id_ = id_[0]
    return id_


@register_graphfcn
class assign_id(BaseGraphFunction, adds_edge_attributes=["id"]):
    """assign_id class."""

    def __call__(self, G: nx.Graph, **kwargs) -> nx.Graph:
        """Assign an ID to each edge.

        This function takes a graph and assigns an ID to each edge. The ID is
        assigned to the 'id' attribute of each edge. Needed because some edges
        have 'osmid', some have a list of 'osmid', others have 'id'.

        Args:
            G (nx.Graph): A graph
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.Graph): The same graph with an ID assigned to each edge
        """
        edge_ids: set[str] = set()
        edges_to_remove = []
        for u, v, key, data in G.edges(data=True, keys=True):
            data["id"] = f"{u}-{v}"
            if data["id"] in edge_ids:
                edges_to_remove.append((u, v, key))
            edge_ids.add(data["id"])
        for u, v, key in edges_to_remove:
            G.remove_edge(u, v, key)
        return G


@register_graphfcn
class remove_parallel_edges(BaseGraphFunction):
    """remove_parallel_edges class."""

    def __call__(self, G: nx.Graph, **kwargs) -> nx.Graph:
        """Remove parallel edges from a street network.

        Retain the edge with the smallest weight (i.e., length).

        Args:
            G (nx.MultiDiGraph): A graph.
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.DiGraph): The graph with parallel edges removed.

        Author:
            Taher Chegini
        """
        # Set the attribute (weight) used to determine which parallel edge to
        # retain. Could make this a parameter in parameters.py if needed.
        weight = "length"
        graph = ox.convert.to_digraph(G)
        _, _, attr_list = next(iter(graph.edges(data=True)))  # type: ignore
        attr_list = cast("dict[str, Any]", attr_list)
        if weight not in attr_list:
            raise ValueError(f"{weight} not in edge attributes.")
        attr = nx.get_node_attributes(graph, weight)
        parallels = (e for e in attr if e[::-1] in attr)
        graph.remove_edges_from(
            {e if attr[e] > attr[e[::-1]] else e[::-1] for e in parallels}
        )
        return graph


@register_graphfcn
class remove_non_pipe_allowable_links(BaseGraphFunction):
    """remove_non_pipe_allowable_links class."""

    def __call__(
        self, G: nx.Graph, topology_derivation: parameters.TopologyDerivation, **kwargs
    ) -> nx.Graph:
        """Remove non-pipe allowable links.

        This function removes links that are not allowable for pipes. The non-
        allowable links are specified in the `omit_edges` attribute of the
        topology_derivation parameter. There two cases handled:

        1. The `highway` property of the edge. In `osmnx`, `highway` is a category
            that contains the road type, e.g., motorway, trunk, primary. If the
            edge contains a value in the `highway` property that is in `omit_edges`,
            the edge is removed.

        2. Any other properties of the edge that are in `omit_edges`. If the
            property is not null in the edge data, the edge is removed. e.g.,
            if `bridge` is in `omit_edges` and the `bridge` entry of the edge
            is NULL, then the edge is retained, if it is something like 'yes',
            or 'viaduct' then the edge is removed.

        Args:
            G (nx.Graph): A graph
            topology_derivation (parameters.TopologyDerivation): A TopologyDerivation
                parameter object
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.Graph): A graph
        """
        edges_to_remove = set()
        for u, v, keys, data in G.edges(data=True, keys=True):
            for omit in topology_derivation.omit_edges:
                if data.get("highway", None) == omit:
                    # Check whether the 'highway' property is 'omit'
                    edges_to_remove.add((u, v, keys))
                elif data.get(omit, None):
                    # Check whether the 'omit' property of edge is not None
                    edges_to_remove.add((u, v, keys))
        for edges in edges_to_remove:
            G.remove_edge(*edges)
        return G


def sum_over_delimiter(s: int | str | float) -> float:
    """Sum over a delimiter.

    This function takes a value, if it is not a string it is casted as a
    float, otherwise it sums over the numbers in the string. The
    numbers are separated by a delimiter. The delimiter is any non-numeric
    character. If the input is not a string, the function returns the input.

    Args:
        s (int | str | float): The input.

    Returns:
        float: The sum of the numbers in the string
    """
    if not isinstance(s, str):
        return float(s)
    return float(sum([int(num) for num in re.split(r"\D+", s) if num]))


@register_graphfcn
class calculate_streetcover(
    BaseGraphFunction, required_edge_attributes=["lanes", "geometry"]
):
    """calculate_streetcover class."""

    # i.e., in osmnx format, i.e., empty for single lane, an int for a
    # number of lanes or a list if the edge has multiple carriageways

    def __call__(
        self,
        G: nx.Graph,
        subcatchment_derivation: parameters.SubcatchmentDerivation,
        addresses: FilePaths,
        **kwargs,
    ) -> nx.Graph:
        """Format the lanes attribute of each edge and calculates width.

        Only the `drive` network is assumed to contribute to impervious area and
        so others `network_types` have lanes set to 0. If no `network_type` is
        present, the edge is assumed to be of type `drive`.

        Args:
            G (nx.Graph): A graph
            subcatchment_derivation (parameters.SubcatchmentDerivation): A
                SubcatchmentDerivation parameter object
            addresses (FilePaths): A FilePaths parameter object
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.Graph): A graph
        """
        G = G.copy()
        lines = []
        for u, v, data in G.edges(data=True):
            if data.get("network_type", "drive") == "drive":
                lanes = data.get("lanes", 1)
            else:
                lanes = 0
            if isinstance(lanes, list):
                lanes = sum([sum_over_delimiter(x) for x in lanes])
            else:
                lanes = sum_over_delimiter(lanes)
            lines.append(
                {
                    "geometry": data["geometry"].buffer(
                        lanes * subcatchment_derivation.lane_width,
                        cap_style=2,
                        join_style=2,
                    ),
                    "u": u,
                    "v": v,
                }
            )
        lines_df = pd.DataFrame(lines)
        lines_gdf = gpd.GeoDataFrame(
            lines_df, geometry=lines_df.geometry, crs=G.graph["crs"]
        )
        if addresses.model_paths.streetcover.suffix in (".geoparquet", ".parquet"):
            lines_gdf.to_parquet(addresses.model_paths.streetcover)
        else:
            lines_gdf.to_file(addresses.model_paths.streetcover, driver="GeoJSON")

        return G


@register_graphfcn
class double_directed(BaseGraphFunction, required_edge_attributes=["id"]):
    """double_directed class."""

    def __call__(self, G: nx.Graph, **kwargs) -> nx.Graph:
        """Create a 'double directed graph'.

        This function duplicates a graph and adds reverse edges to the new graph.
        These new edges share the same data as the 'forward' edges but have a new
        'id'. An undirected graph is not suitable because it removes data of one of
        the edges if there are edges in both directions between two nodes
        (necessary to preserve, e.g., consistent 'width'). If 'edge_type' is
        present, then the function will only be performed on 'street' types.

        Args:
            G (nx.Graph): A graph
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.Graph): A graph
        """
        # Convert to directed
        G_new = nx.MultiDiGraph(G.copy())

        # MultiDiGraph adds edges in both directions, but rivers (and geometries)
        # are only in one direction. So we remove the reverse edges and add them
        # back in with the correct geometry.
        # This assumes that 'id' is of format 'start-end' (see assign_id)
        arcs_to_remove = [
            (u, v) for u, v, d in G_new.edges(data=True) if f"{u}-{v}" != d.get("id")
        ]

        # Remove the reverse edges
        for u, v in arcs_to_remove:
            G_new.remove_edge(u, v)

        # Add in reversed edges for streets only and with geometry
        for u, v, data in G.edges(data=True):
            include = data.get("edge_type", True)
            if isinstance(include, str):
                include = include == "street"
            if ((v, u) not in G_new.edges) & include:
                reverse_data = data.copy()
                reverse_data["id"] = f"{data['id']}.reversed"
                new_geometry = shapely.LineString(
                    list(reversed(data["geometry"].coords))
                )
                reverse_data["geometry"] = new_geometry
                G_new.add_edge(v, u, **reverse_data)
        return G_new


@register_graphfcn
class to_undirected(BaseGraphFunction):
    """to_undirected class."""

    def __call__(self, G: nx.Graph, **kwargs) -> nx.Graph:
        """Convert the graph to an undirected graph.

        Args:
            G (nx.Graph): A graph
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.Graph): An undirected graph
        """
        # Don't use osmnx.to_undirected! It enables multigraph if the geometries
        # are different, but we have already saved the street cover so don't
        # want this!
        return G.to_undirected()


@register_graphfcn
class split_long_edges(BaseGraphFunction, required_edge_attributes=["id", "geometry"]):
    """split_long_edges class."""

    def __call__(
        self,
        G: nx.Graph,
        subcatchment_derivation: parameters.SubcatchmentDerivation,
        **kwargs,
    ) -> nx.Graph:
        """Split long edges into shorter edges.

        This function splits long edges into shorter edges. The edges are split
        into segments of length 'max_street_length'. The 'geometry' of the
        original edge must be a LineString. Intended to follow up with call of
        `merge_nodes`.

        Args:
            G (nx.Graph): A graph
            subcatchment_derivation (parameters.SubcatchmentDerivation): A
                SubcatchmentDerivation parameter object
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            graph (nx.Graph): A graph
        """
        max_length = subcatchment_derivation.max_street_length

        # Split edges
        new_linestrings = shapely.segmentize(
            [d["geometry"] for u, v, d in G.edges(data=True)], max_length
        )
        new_nodes = shapely.get_coordinates(new_linestrings)

        new_edges = {}
        for new_linestring, (u, v, d) in zip(new_linestrings, G.edges(data=True)):
            # Create an arc for each segment
            for start, end in zip(
                new_linestring.coords[:-1], new_linestring.coords[1:]
            ):
                geom = shapely.LineString([start, end])
                new_edges[(start, end, 0)] = {**d, "length": geom.length}

        # Create new graph
        new_graph = nx.MultiGraph()
        new_graph.graph = G.graph.copy()
        new_graph.add_edges_from(new_edges)
        nx.set_edge_attributes(new_graph, new_edges)
        nx.set_node_attributes(
            new_graph, {tuple(node): {"x": node[0], "y": node[1]} for node in new_nodes}
        )
        return nx.relabel_nodes(
            new_graph, {node: ix for ix, node in enumerate(new_graph.nodes)}
        )


@register_graphfcn
class merge_street_nodes(BaseGraphFunction):
    """merge_nodes class."""

    def __call__(
        self,
        G: nx.Graph,
        subcatchment_derivation: parameters.SubcatchmentDerivation,
        **kwargs,
    ) -> nx.Graph:
        """Merge nodes that are close together.

        Merges `street` nodes that are within a certain distance of each
        other. The distance is specified in the `node_merge_distance` attribute
        of the `subcatchment_derivation` parameter. The merged nodes are given
        the same coordinates, and the graph is relabeled with nx.relabel_nodes.
        Suggest to follow with call of `assign_id` to remove duplicate edges.

        Args:
            G (nx.Graph): A graph
            subcatchment_derivation (parameters.SubcatchmentDerivation): A
                SubcatchmentDerivation parameter object
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.Graph): A graph
        """
        G = G.copy()

        # Separate out streets
        street_edges = [
            (u, v, k)
            for u, v, k, d in G.edges(data=True, keys=True)
            if d.get("edge_type", "street") == "street"
        ]
        streets = G.edge_subgraph(street_edges).copy()

        # Identify nodes that are within threshold of each other
        mapping = go.merge_points(
            [(d["x"], d["y"]) for u, d in streets.nodes(data=True)],
            subcatchment_derivation.node_merge_distance,
        )

        # Get indexes of node names
        node_indices = {ix: node for ix, node in enumerate(streets.nodes)}

        # Create a mapping of old node names to new node names
        node_names = {}
        for ix, node in enumerate(streets.nodes):
            if ix in mapping:
                # If the node is in the mapping, then it is mapped and
                # given the new coordinate (all nodes in a mapping family must
                # be given the same coordinate because of how relabel_nodes
                # works)
                node_names[node] = node_indices[mapping[ix]["maps_to"]]
                G.nodes[node]["x"] = mapping[ix]["coordinate"][0]
                G.nodes[node]["y"] = mapping[ix]["coordinate"][1]
            else:
                node_names[node] = node

        G = nx.relabel_nodes(G, node_names)

        # Relabelling will create selfloops within a mapping family, which
        # are removed
        self_loops = list(nx.selfloop_edges(G))
        G.remove_edges_from(self_loops)

        return G


@register_graphfcn
class fix_geometries(
    BaseGraphFunction,
    adds_edge_attributes=["geometry"],
    required_node_attributes=["x", "y"],
):
    """fix_geometries class."""

    def __call__(self, G: nx.Graph, **kwargs) -> nx.Graph:
        """Fix the geometries of the edges.

        This function fixes the geometries of the edges. The geometries are
        recalculated from the node coordinates.

        Args:
            G (nx.Graph): A graph
            **kwargs: Additional keyword arguments are ignored.

        Returns:
            G (nx.Graph): A graph
        """
        G = G.copy()
        for u, v, data in G.edges(data=True):
            geom = data.get("geometry", None)

            start_point_node = (G.nodes[u]["x"], G.nodes[u]["y"])
            end_point_node = (G.nodes[v]["x"], G.nodes[v]["y"])
            if not geom:
                start_point_edge = (None, None)
                end_point_edge = (None, None)
            else:
                start_point_edge = data["geometry"].coords[0]
                end_point_edge = data["geometry"].coords[-1]

            if (start_point_edge == end_point_node) & (
                end_point_edge == start_point_node
            ):
                data["geometry"] = data["geometry"].reverse()
            elif (start_point_edge != start_point_node) | (
                end_point_edge != end_point_node
            ):
                data["geometry"] = shapely.LineString(
                    [start_point_node, end_point_node]
                )
        return G



"""The entry point for the swmmanywhere program."""

from __future__ import annotations

import argparse
import os
from pathlib import Path

from swmmanywhere import swmmanywhere
from swmmanywhere.logging import logger


def parse_arguments():
    """Parse the command line arguments.

    Returns:
        Path: The path to the configuration file.
    """
    parser = argparse.ArgumentParser(description="Process command line arguments.")
    parser.add_argument(
        "--config_path",
        type=Path,
        default=Path(__file__).parent.parent
        / "tests"
        / "test_data"
        / "demo_config.yml",
        help="Configuration file path",
    )
    parser.add_argument(
        "--verbose", type=bool, default=False, help="Configuration verbosity"
    )
    args = parser.parse_args()
    return args.config_path, args.verbose


def run():
    """Run a swmmanywhere config file."""
    # Parse the arguments
    config_path, verbose = parse_arguments()
    os.environ["SWMMANYWHERE_VERBOSE"] = str(verbose).lower()

    config = swmmanywhere.load_config(config_path)

    # Run the model
    inp, metrics = swmmanywhere.swmmanywhere(config)
    logger.info(f"Model run complete. Results saved to {inp}")
    logger.info(f"Metrics:\n {metrics}")


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
"""Created on 2024-01-26.

@author: Barney
"""

from __future__ import annotations

import math
import os
import tempfile
from pathlib import Path

import geopandas as gpd
import networkx as nx
import pytest
from shapely import geometry as sgeom

from swmmanywhere import parameters
from swmmanywhere.filepaths import FilePaths
from swmmanywhere.graph_utilities import (
    filter_streets,
    iterate_graphfcns,
    load_graph,
    save_graph,
    validate_graphfcn_list,
)
from swmmanywhere.graph_utilities import graphfcns as gu
from swmmanywhere.logging import logger


@pytest.fixture
def street_network():
    """Load a street network."""
    bbox = (-0.11643, 51.50309, -0.11169, 51.50549)
    G = load_graph(Path(__file__).parent / "test_data" / "street_graph.json")
    return G, bbox


def test_save_load(street_network):
    """Test the save_graph and load_graph functions."""
    G, _ = street_network
    # Load a street network
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        # Save the graph
        save_graph(G, Path(temp_dir) / "test_graph.json")
        # Load the graph
        G_new = load_graph(Path(temp_dir) / "test_graph.json")
        # Check if the loaded graph is the same as the original graph
        assert nx.is_isomorphic(G, G_new)


def test_assign_id(street_network):
    """Test the assign_id function."""
    G, _ = street_network
    G = gu.assign_id(G)
    for u, v, data in G.edges(data=True):
        assert "id" in data.keys()
        assert isinstance(data["id"], str)


def test_double_directed(street_network):
    """Test the double_directed function."""
    G, _ = street_network
    G = gu.assign_id(G)
    G = gu.double_directed(G)
    for u, v in G.edges():
        assert (v, u) in G.edges


def test_calculate_streetcover(street_network):
    """Test the calculate_streetcover function."""
    G, _ = street_network
    params = parameters.SubcatchmentDerivation()
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        addresses = FilePaths(
            base_dir=Path(temp_dir),
            bbox_bounds=[0, 1, 0, 1],
            project_name="",
            extension="json",
            streetcover=Path(temp_dir) / "streetcover.geojson",
        )
        _ = gu.calculate_streetcover(G, params, addresses)
        # TODO test that G hasn't changed? or is that a waste of time?
        assert addresses.model_paths.streetcover.exists()
        gdf = gpd.read_file(addresses.model_paths.streetcover)
        assert len(gdf) == len(G.edges)
        assert gdf.geometry.area.sum() > 0

        # Test odd lanes formatting
        """Test the streetcover function with oddly formatted lanes."""
        nx.set_edge_attributes(G, "2;1", "lanes")
        _ = gu.calculate_streetcover(G, params, addresses)
        assert addresses.model_paths.streetcover.exists()
        gdf1 = gpd.read_file(addresses.model_paths.streetcover)

        nx.set_edge_attributes(G, 3, "lanes")
        _ = gu.calculate_streetcover(G, params, addresses)
        assert addresses.model_paths.streetcover.exists()
        gdf2 = gpd.read_file(addresses.model_paths.streetcover)

        assert gdf1.geometry.area.sum() == gdf2.geometry.area.sum()


def test_split_long_edges(street_network):
    """Test the split_long_edges function."""
    G, _ = street_network
    G = gu.assign_id(G)
    max_length = 40
    params = parameters.SubcatchmentDerivation(max_street_length=max_length)
    G = gu.split_long_edges(G, params)
    for u, v, data in G.edges(data=True):
        assert data["length"] <= (max_length * 2)


def test_derive_subcatchments(street_network):
    """Test the derive_subcatchments function."""
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        temp_path = Path(temp_dir)
        addresses = FilePaths(
            base_dir=Path(temp_dir),
            bbox_bounds=[0, 1, 0, 1],
            project_name="test",
            extension="json",
            elevation=Path(__file__).parent / "test_data" / "elevation.tif",
            building=temp_path / "building.geojson",
            streetcover=temp_path / "building.geojson",
            subcatchments=temp_path / "subcatchments.geojson",
        )
        params = parameters.SubcatchmentDerivation()
        G, _ = street_network

        # mock up buildings
        eg_bldg = sgeom.Polygon(
            [
                (700291.346, 5709928.922),
                (700331.206, 5709927.815),
                (700321.610, 5709896.444),
                (700293.192, 5709900.503),
                (700291.346, 5709928.922),
            ]
        )
        gdf = gpd.GeoDataFrame(geometry=[eg_bldg], crs=G.graph["crs"])
        gdf.to_file(addresses.bbox_paths.building, driver="GeoJSON")

        G = gu.calculate_contributing_area(G, params, addresses)
        for u, v, data in G.edges(data=True):
            assert "contributing_area" in data.keys()
            assert isinstance(data["contributing_area"], float)

        for u, data in G.nodes(data=True):
            assert "contributing_area" in data.keys()
            assert isinstance(data["contributing_area"], float)


def test_set_elevation_and_slope(street_network):
    """Test the set_elevation, set_surface_slope, chahinian_slope function."""
    G, _ = street_network
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        addresses = FilePaths(
            base_dir=Path(temp_dir),
            bbox_bounds=[0, 1, 0, 1],
            project_name="test",
            extension="json",
            elevation=Path(__file__).parent / "test_data" / "elevation.tif",
        )
        G = gu.set_elevation(G, addresses)
        for id_, data in G.nodes(data=True):
            assert "surface_elevation" in data.keys()
            assert math.isfinite(data["surface_elevation"])
            assert data["surface_elevation"] > 0

        G = gu.set_surface_slope(G)
        for u, v, data in G.edges(data=True):
            assert "surface_slope" in data.keys()
            assert math.isfinite(data["surface_slope"])

        G = gu.set_chahinian_slope(G)
        for u, v, data in G.edges(data=True):
            assert "chahinian_slope" in data.keys()
            assert math.isfinite(data["chahinian_slope"])

        slope_vals = {-2: 1, 0.3: 0, 0.4: 0, 12: 1}
        for slope, expected in slope_vals.items():
            first_edge = list(G.edges)[0]
            G.edges[first_edge]["surface_slope"] = slope / 100
            G = gu.set_chahinian_slope(G)
            assert G.edges[first_edge]["chahinian_slope"] == expected


def test_chahinian_angle(street_network):
    """Test the chahinian_angle function."""
    G, _ = street_network
    G = gu.set_chahinian_angle(G)
    for u, v, data in G.edges(data=True):
        assert "chahinian_angle" in data.keys()
        assert math.isfinite(data["chahinian_angle"])


def test_calculate_weights(street_network):
    """Test the calculate_weights function."""
    G, _ = street_network
    params = parameters.TopologyDerivation()
    for weight in params.weights:
        for ix, (u, v, data) in enumerate(G.edges(data=True)):
            data[weight] = ix

    G = gu.calculate_weights(G, params)
    for u, v, data in G.edges(data=True):
        assert "weight" in data.keys()
        assert math.isfinite(data["weight"])


def test_calculate_weights_novar(street_network):
    """Test the calculate_weights function with no variance."""
    G, _ = street_network
    params = parameters.TopologyDerivation()
    for weight in params.weights:
        for ix, (u, v, data) in enumerate(G.edges(data=True)):
            data[weight] = 1.5

    G = gu.calculate_weights(G, params)
    for u, v, data in G.edges(data=True):
        assert "weight" in data.keys()
        assert math.isfinite(data["weight"])


def test_identify_outfalls_no_river(street_network):
    """Test the identify_outfalls in the no river case."""
    G, _ = street_network
    G = gu.assign_id(G)
    G = gu.double_directed(G)
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        addresses = FilePaths(
            base_dir=Path(temp_dir),
            bbox_bounds=[0, 1, 0, 1],
            project_name="test",
            extension="json",
            elevation=Path(__file__).parent / "test_data" / "elevation.tif",
        )
        G = gu.set_elevation(G, addresses)
        for ix, (u, v, d) in enumerate(G.edges(data=True)):
            d["edge_type"] = "street"
            d["weight"] = ix
        params = parameters.OutfallDerivation()
        G = gu.identify_outfalls(G, params)
        outfalls = [
            (u, v, d) for u, v, d in G.edges(data=True) if d["edge_type"] == "outfall"
        ]
        assert len(outfalls) == 1


def test_identify_outfalls_sg(street_network):
    """Test the identify_outfalls with subgraphs."""
    G, _ = street_network

    G = gu.assign_id(G)
    G = gu.double_directed(G)
    elev_fid = Path(__file__).parent / "test_data" / "elevation.tif"
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        addresses = FilePaths(
            base_dir=Path(temp_dir),
            bbox_bounds=[0, 1, 0, 1],
            project_name="test",
            extension="json",
            elevation=elev_fid,
        )
        G = gu.set_elevation(G, addresses)
        for ix, (u, v, d) in enumerate(G.edges(data=True)):
            d["edge_type"] = "street"
            d["weight"] = ix

        params = parameters.OutfallDerivation(
            river_buffer_distance=200, outfall_length=10, method="withtopo"
        )
        dummy_river1 = sgeom.LineString(
            [(699913.878, 5709769.851), (699932.546, 5709882.575)]
        )
        dummy_river2 = sgeom.LineString(
            [(699932.546, 5709882.575), (700011.524, 5710060.636)]
        )
        dummy_river3 = sgeom.LineString(
            [(700011.524, 5710060.636), (700103.427, 5710169.052)]
        )

        G.add_edge(
            "river1",
            "river2",
            **{
                "length": 10,
                "edge_type": "river",
                "id": "river1-to-river2",
                "geometry": dummy_river1,
            },
        )
        G.add_edge(
            "river2",
            "river3",
            **{
                "length": 10,
                "edge_type": "river",
                "id": "river2-to-river3",
                "geometry": dummy_river2,
            },
        )

        G.add_edge(
            "river3",
            "river4",
            **{
                "length": 10,
                "edge_type": "river",
                "id": "river3-to-river4",
                "geometry": dummy_river3,
            },
        )

        G.nodes["river1"]["x"] = 699913.878
        G.nodes["river1"]["y"] = 5709769.851
        G.nodes["river2"]["x"] = 699932.546
        G.nodes["river2"]["y"] = 5709882.575
        G.nodes["river3"]["x"] = 700011.524
        G.nodes["river3"]["y"] = 5710060.636
        G.nodes["river4"]["x"] = 700103.427
        G.nodes["river4"]["y"] = 5710169.052

        # Cut into subgraphs
        G.remove_edge(12354833, 25472373)
        G.remove_edge(25472373, 12354833)
        G.remove_edge(109753, 25472854)
        G.remove_edge(25472854, 109753)

        # Test outfall derivation
        G_ = G.copy()
        G_ = gu.identify_outfalls(G_, params)

        # Two subgraphs = two routes to waste
        outfalls = [
            (u, v, d)
            for u, v, d in G_.edges(data=True)
            if d["edge_type"] == "waste-outfall"
        ]
        assert len(outfalls) == 2

        # With buffer distance 300, the subgraph near the river will have an outfall
        # between the nearest street node to each river node (there are 3 potential
        # links in 150m). The subgraph further from the river is too far to be linked
        # to the river nodes and so will have a dummy river node as an outfall. 3+1=5
        outfalls = [
            (u, v, d) for u, v, d in G_.edges(data=True) if d["edge_type"] == "outfall"
        ]
        assert len(outfalls) == 3


def test_identify_outfalls_and_derive_topology(street_network):
    """Test the identify_outfalls and derive_topology functions."""
    G, _ = street_network
    G = gu.assign_id(G)
    G = gu.double_directed(G)
    for ix, (u, v, d) in enumerate(G.edges(data=True)):
        d["edge_type"] = "street"
        d["weight"] = ix

    params = parameters.OutfallDerivation(
        river_buffer_distance=200, outfall_length=10, method="separate"
    )
    dummy_river1 = sgeom.LineString(
        [(699913.878, 5709769.851), (699932.546, 5709882.575)]
    )
    dummy_river2 = sgeom.LineString(
        [(699932.546, 5709882.575), (700011.524, 5710060.636)]
    )
    dummy_river3 = sgeom.LineString(
        [(700011.524, 5710060.636), (700103.427, 5710169.052)]
    )

    G.add_edge(
        "river1",
        "river2",
        **{
            "length": 10,
            "edge_type": "river",
            "id": "river1-to-river2",
            "geometry": dummy_river1,
        },
    )
    G.add_edge(
        "river2",
        "river3",
        **{
            "length": 10,
            "edge_type": "river",
            "id": "river2-to-river3",
            "geometry": dummy_river2,
        },
    )

    G.add_edge(
        "river3",
        "river4",
        **{
            "length": 10,
            "edge_type": "river",
            "id": "river3-to-river4",
            "geometry": dummy_river3,
        },
    )

    G.nodes["river1"]["x"] = 699913.878
    G.nodes["river1"]["y"] = 5709769.851
    G.nodes["river2"]["x"] = 699932.546
    G.nodes["river2"]["y"] = 5709882.575
    G.nodes["river3"]["x"] = 700011.524
    G.nodes["river3"]["y"] = 5710060.636
    G.nodes["river4"]["x"] = 700103.427
    G.nodes["river4"]["y"] = 5710169.052

    # Test outfall derivation
    G_ = G.copy()
    G_ = gu.identify_outfalls(G_, params)

    outfalls = [
        (u, v, d) for u, v, d in G_.edges(data=True) if d["edge_type"] == "outfall"
    ]
    assert len(outfalls) == 2

    # Test topo derivation
    G_ = gu.derive_topology(G_, params)
    assert len(G_.edges) == 22
    assert len(set([d["outfall"] for u, d in G_.nodes(data=True)])) == 2
    for u, d in G_.nodes(data=True):
        assert "x" in d.keys()
        assert "y" in d.keys()

    # Test outfall derivation parameters
    G_ = G.copy()
    params.outfall_length = 600
    G_ = gu.identify_outfalls(G_, params)
    outfalls = [
        (u, v, d) for u, v, d in G_.edges(data=True) if d["edge_type"] == "outfall"
    ]
    assert len(outfalls) == 1


def test_identify_outfalls_and_derive_topology_withtopo(street_network):
    """Test the identify_outfalls and derive_topology functions."""
    G, _ = street_network
    G = gu.assign_id(G)
    G = gu.double_directed(G)
    for ix, (u, v, d) in enumerate(G.edges(data=True)):
        d["edge_type"] = "street"
        d["weight"] = ix

    params = parameters.OutfallDerivation(
        river_buffer_distance=250, outfall_length=10, method="withtopo"
    )
    dummy_river1 = sgeom.LineString(
        [(699913.878, 5709769.851), (699932.546, 5709882.575)]
    )
    dummy_river2 = sgeom.LineString(
        [(699932.546, 5709882.575), (700011.524, 5710060.636)]
    )
    dummy_river3 = sgeom.LineString(
        [(700011.524, 5710060.636), (700103.427, 5710169.052)]
    )

    G.add_edge(
        "river1",
        "river2",
        **{
            "length": 10,
            "edge_type": "river",
            "id": "river1-to-river2",
            "geometry": dummy_river1,
        },
    )
    G.add_edge(
        "river2",
        "river3",
        **{
            "length": 10,
            "edge_type": "river",
            "id": "river2-to-river3",
            "geometry": dummy_river2,
        },
    )

    G.add_edge(
        "river3",
        "river4",
        **{
            "length": 10,
            "edge_type": "river",
            "id": "river3-to-river4",
            "geometry": dummy_river3,
        },
    )

    G.nodes["river1"]["x"] = 699913.878
    G.nodes["river1"]["y"] = 5709769.851
    G.nodes["river2"]["x"] = 699932.546
    G.nodes["river2"]["y"] = 5709882.575
    G.nodes["river3"]["x"] = 700011.524
    G.nodes["river3"]["y"] = 5710060.636
    G.nodes["river4"]["x"] = 700103.427
    G.nodes["river4"]["y"] = 5710169.052

    # Test outfall derivation
    G_ = G.copy()
    G_ = gu.identify_outfalls(G_, params)

    outfalls = [
        (u, v, d)
        for u, v, d in G_.edges(data=True)
        if d["edge_type"] == "waste-outfall"
    ]
    assert len(outfalls) == 1

    outfalls = [
        (u, v, d) for u, v, d in G_.edges(data=True) if d["edge_type"] == "outfall"
    ]
    assert len(outfalls) == 4

    # Test topo derivation
    G_ = gu.derive_topology(G_, params)
    assert len(G_.edges) == 20
    assert len(set([d["outfall"] for u, d in G_.nodes(data=True)])) == 4

    # Test outfall derivation parameters
    G_ = G.copy()
    params.outfall_length = 600
    G_ = gu.identify_outfalls(G_, params)
    G_ = gu.derive_topology(G_, params)
    assert len(set([d["outfall"] for u, d in G_.nodes(data=True)])) == 1
    for u, d in G_.nodes(data=True):
        assert "x" in d.keys()
        assert "y" in d.keys()


def test_pipe_by_pipe():
    """Test the pipe_by_pipe function."""
    G = load_graph(Path(__file__).parent / "test_data" / "graph_topo_derived.json")
    for ix, (u, d) in enumerate(G.nodes(data=True)):
        d["surface_elevation"] = ix
        d["contributing_area"] = ix

    # 107738 and 107734 join to 107733. Since 107738 has fewer ancestors, it is
    # processed first. We artificially set it to a lower elevation than 107734 to ensure
    # that the design of 107734 does not overwrite this lower required depth from 107738
    G.nodes[107738]["surface_elevation"] = -1

    params = parameters.HydraulicDesign()

    G = gu.pipe_by_pipe(G, params)
    for u, v, d in G.edges(data=True):
        assert "diameter" in d.keys()
        assert "cost_usd" in d.keys()
        assert d["diameter"] in params.diameters

    for u, d in G.nodes(data=True):
        assert "chamber_floor_elevation" in d.keys()
        assert math.isfinite(d["chamber_floor_elevation"])

    assert G.nodes[107733]["chamber_floor_elevation"] < -1


def get_edge_types(G):
    """Get the edge types in the graph."""
    edge_types = set()
    for u, v, d in G.edges(data=True):
        if isinstance(d["highway"], list):
            edge_types.union(d["highway"])
        else:
            edge_types.add(d["highway"])
    return edge_types


def test_remove_non_pipe_allowable_links():
    """Test the remove_non_pipe_allowable_links function."""
    G = load_graph(Path(__file__).parent / "test_data" / "street_graph.json")
    # Ensure some invalid paths
    topology_params = parameters.TopologyDerivation(omit_edges=["primary", "bridge"])

    # Test that an edge has a non-None 'bridge' entry
    assert len(set([d.get("bridge", None) for u, v, d in G.edges(data=True)])) > 1

    # Test that an edge has a 'primary' entry under highway
    assert "primary" in get_edge_types(G)

    G_ = gu.remove_non_pipe_allowable_links(G, topology_params)
    assert "primary" not in get_edge_types(G_)
    assert len(set([d.get("bridge", None) for u, v, d in G_.edges(data=True)])) == 1


def test_iterate_graphfcns():
    """Test the iterate_graphfcns function."""
    G = load_graph(Path(__file__).parent / "test_data" / "graph_topo_derived.json")
    params = parameters.get_full_parameters()
    params["topology_derivation"].omit_edges = ["primary", "bridge"]
    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        temp_path = Path(temp_dir)
        addresses = FilePaths(
            base_dir=temp_path,
            bbox_bounds=[0, 1, 0, 1],
            project_name="test",
            extension="json",
            model=temp_path,
        )

        G = iterate_graphfcns(
            G, ["assign_id", "remove_non_pipe_allowable_links"], params, addresses
        )
        for u, v, d in G.edges(data=True):
            assert "id" in d.keys()
        assert "primary" not in get_edge_types(G)
        assert len(set([d.get("bridge", None) for u, v, d in G.edges(data=True)])) == 1


def _remove_edges(G: nx.Graph, **kw):
    """Remove all edges from the graph.

    Args:
        G (nx.Graph): The graph to remove edges from.
        kw: Additional keyword arguments. (which are ignored)

    Returns:
        nx.Graph: The graph with no edges.
    """
    G.remove_edges_from(list(G.edges))
    return G


def test_iterate_graphfcns_noedges():
    """Test the iterate_graphfcns function for a graph with no edges."""
    G = load_graph(Path(__file__).parent / "test_data" / "graph_topo_derived.json")

    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        temp_path = Path(temp_dir)
        addresses = FilePaths(
            base_dir=temp_path,
            bbox_bounds=[0, 1, 0, 1],
            project_name="test",
            extension="json",
            model=temp_path,
        )
        os.environ["SWMMANYWHERE_VERBOSE"] = "true"
        original_function = gu["remove_non_pipe_allowable_links"]
        gu["remove_non_pipe_allowable_links"] = _remove_edges
        G = iterate_graphfcns(
            G, ["assign_id", "remove_non_pipe_allowable_links"], {}, addresses
        )
        gu["remove_non_pipe_allowable_links"] = original_function
        os.environ["SWMMANYWHERE_VERBOSE"] = "false"
        assert (addresses.model_paths.model / "assign_id_graph.json").exists()
        assert not (
            addresses.model_paths.model / "remove_non_pipe_allowable_links_graph.json"
        ).exists()


def test_fix_geometries():
    """Test the fix_geometries function."""
    # Create a graph with edge geometry not matching node coordinates
    G = load_graph(Path(__file__).parent / "test_data" / "graph_topo_derived.json")

    # Test doesn't work if this isn't true
    assert G.get_edge_data(107733, 25472373, 0)["geometry"].coords[0] != (
        G.nodes[107733]["x"],
        G.nodes[107733]["y"],
    )

    # Run the function
    G_fixed = gu.fix_geometries(G)

    # Check that the edge geometry now matches the node coordinates
    assert G_fixed.get_edge_data(107733, 25472373, 0)["geometry"].coords[0] == (
        G_fixed.nodes[107733]["x"],
        G_fixed.nodes[107733]["y"],
    )


def almost_equal(a, b, tol=1e-6):
    """Check if two numbers are almost equal."""
    return abs(a - b) < tol


def test_merge_street_nodes(street_network):
    """Test the merge_street_nodes function."""
    G, _ = street_network
    subcatchment_derivation = parameters.SubcatchmentDerivation(node_merge_distance=20)
    G_ = gu.merge_street_nodes(G, subcatchment_derivation)
    assert not set([107736, 266325461, 2623975694, 32925453]).intersection(G_.nodes)
    assert almost_equal(G_.nodes[25510321]["x"], 700445.0112082)


def test_clip_to_catchments(street_network):
    """Test the clip_to_catchments function."""
    G, _ = street_network

    with tempfile.TemporaryDirectory(dir=".") as temp_dir:
        os.environ["SWMMANYWHERE_VERBOSE"] = "true"
        temp_path = Path(temp_dir)
        addresses = FilePaths(
            base_dir=temp_path,
            bbox_bounds=[0, 1, 0, 1],
            project_name="test",
            extension="json",
            nodes=temp_path / "nodes.geojson",
            edges=temp_path / "edges.geojson",
            elevation=Path(__file__).parent / "test_data" / "elevation.tif",
        )

        # Test default clipping
        subcatchment_derivation = parameters.SubcatchmentDerivation()
        G_ = gu.clip_to_catchments(
            G, addresses=addresses, subcatchment_derivation=subcatchment_derivation
        )
        assert len(G_.edges) == 7

        # Test default clipping streamorder
        subcatchment_derivation = parameters.SubcatchmentDerivation()
        subcatchment_derivation.subbasin_streamorder = 2
        G_ = gu.clip_to_catchments(
            G, addresses=addresses, subcatchment_derivation=subcatchment_derivation
        )
        assert len(G_.edges) == 3

        # Test clipping
        subcatchment_derivation = parameters.SubcatchmentDerivation(
            subbasin_streamorder=3,
            subbasin_membership=0.9,
            subbasin_clip_method="community",
        )
        G_ = gu.clip_to_catchments(
            G, addresses=addresses, subcatchment_derivation=subcatchment_derivation
        )
        assert len(G_.edges) == 30

        # Test clipping with different params
        subcatchment_derivation = parameters.SubcatchmentDerivation(
            subbasin_streamorder=4,
            subbasin_membership=0.3,
            subbasin_clip_method="community",
        )
        G_ = gu.clip_to_catchments(
            G, addresses=addresses, subcatchment_derivation=subcatchment_derivation
        )
        assert len(G_.edges) == 34

        # Test no cuts
        subcatchment_derivation = parameters.SubcatchmentDerivation(
            subbasin_streamorder=4,
            subbasin_membership=0,
            subbasin_clip_method="community",
        )
        G_ = gu.clip_to_catchments(
            G, addresses=addresses, subcatchment_derivation=subcatchment_derivation
        )
        assert len(G_.edges) == 39

        # Cut between every community not entirely within the same basin
        subcatchment_derivation = parameters.SubcatchmentDerivation(
            subbasin_streamorder=4,
            subbasin_membership=1,
            subbasin_clip_method="community",
        )
        G_ = gu.clip_to_catchments(
            G, addresses=addresses, subcatchment_derivation=subcatchment_derivation
        )
        assert len(G_.edges) == 30

        # Check streamorder adjustment
        with tempfile.NamedTemporaryFile(
            suffix=".log", mode="w+b", delete=False
        ) as temp_file:
            fid = Path(temp_file.name)
            os.environ["SWMMANYWHERE_VERBOSE"] = "true"
            logger.add(fid)
            subcatchment_derivation = parameters.SubcatchmentDerivation(
                subbasin_streamorder=5, subbasin_membership=0.9
            )
            G_ = gu.clip_to_catchments(
                G, addresses=addresses, subcatchment_derivation=subcatchment_derivation
            )
            ftext = str(temp_file.read())
            assert """No subbasins found""" in ftext
            assert """WARNING""" in ftext
            logger.remove()
            os.environ["SWMMANYWHERE_VERBOSE"] = "false"
        assert (addresses.model_paths.nodes.parent / "subbasins.geojson").exists()


def test_filter_streets():
    """Test the _filter_streets function."""
    # Create a sample graph
    G = nx.Graph()
    G.add_edges_from(
        [
            (1, 2, {"edge_type": "street"}),
            (2, 3, {"edge_type": "street"}),
            (3, 4, {"edge_type": "outfall"}),
            (4, 5, {"edge_type": "river"}),
        ]
    )

    # Test case 1: Filter streets
    G_streets = filter_streets(G)
    assert set(G_streets.nodes) == {1, 2, 3}
    assert set(G_streets.edges) == {(1, 2), (2, 3)}

    # Test case 2: Empty graph
    G_empty = nx.Graph()
    G_empty_streets = filter_streets(G_empty)
    assert len(G_empty_streets.nodes) == 0
    assert len(G_empty_streets.edges) == 0

    # Test case 3: All non-street edges
    G_non_streets = nx.Graph()
    G_non_streets.add_edges_from(
        [(1, 2, {"edge_type": "non-street"}), (2, 3, {"edge_type": "non-street"})]
    )
    G_non_streets_filtered = filter_streets(G_non_streets)
    assert len(G_non_streets_filtered.nodes) == 0
    assert len(G_non_streets_filtered.edges) == 0


def test_validate_graphfcn_list(street_network):
    """Test the validate_graphfcn_list function."""
    # Test case 1: Valid list
    validate_graphfcn_list(["assign_id", "double_directed"])

    # Test case 2: Invalid list
    with pytest.raises(ValueError) as exc_info:
        validate_graphfcn_list(["assign_id", "not_a_function"])
    assert "not_a_function" in str(exc_info.value)

    # Test case 3: Valid order
    G, _ = street_network
    validate_graphfcn_list(["assign_id", "double_directed"], G)

    # Test case 4: Invalid order
    with pytest.raises(ValueError) as exc_info:
        validate_graphfcn_list(["assign_id", "calculate_weights"], G)
    assert "calculate_weights requires edge attributes" in str(exc_info.value)
    assert "chahinian_angle" in str(exc_info.value)



from __future__ import annotations

from swmmanywhere import parameters


@parameters.register_parameter_group(name="new_params")
class new_params(parameters.BaseModel):
    """New parameters."""

    new_param: int = parameters.Field(
        default=1,
        ge=0,
        le=10,
        unit="-",
        description="A new parameter.",
    )



"""The main module for MyProject."""

from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("swmmanywhere")
except PackageNotFoundError:
    __version__ = "999"

# Importing module to register the graphfcns and made them available
from . import graphfcns  # noqa: F401
